{% extends "default/index.html.jinja2" %}
<style>
    {% block style_theme %}
        :root {
            --pdoc-background: #272935;
        }
        .pdoc {
            --text: #FFF;
            --accent: #333;
            --active: #9e7700;
        }
    {% endblock %}
</style>
