pre {
    line-height: 125%;
}

td.linenos pre {
    color: #ff5555;
    background-color: #282a36;
    padding-left: 5px;
    padding-right: 5px;
}

span.linenos {
    color: #ff5555;
    background-color: #282a36;
    padding-left: 5px;
    padding-right: 5px;
}

td.linenos pre.special {
    color: #ff5555;
    background-color: #ffffc0;
    padding-left: 5px;
    padding-right: 5px;
}

span.linenos.special {
    color: #ff5555;
    background-color: #ffffc0;
    padding-left: 5px;
    padding-right: 5px;
}

/* This is for the source code docs */

.pdoc .hll {
    background-color: #282a36;
}

.pdoc {
    background: #282a36;
    color: #ff5555;
}

/* Comment */

.pdoc .c {
    color: #6a7aaa;
}

/* Error */
.pdoc .err {
    color: #ff5555;
    background-color: #1e0010;
}

/* Keyword */
.pdoc .k {
    color: #ff79c6;
}

/* Literal */
.pdoc .l {
    color: #ae81ff;
}

/* Name */
.pdoc .n {
    color: #f8f8f2;
}

/* Operator */
.pdoc .o {
    color: #ff79c6;
}

/* Punctuation */
.pdoc .p {
    color: #f8f8f2;
}

/* Comment.Hashbang */
.pdoc .ch {
    color: #6a7aaa;
}

/* Comment.Multiline */
.pdoc .cm {
    color: #6a7aaa;
}

/* Comment.Preproc */
.pdoc .cp {
    color: #6a7aaa;
}

/* Comment.PreprocFile */
.pdoc .cpf {
    color: #6a7aaa;
}

/* Comment.Single */
.pdoc .c1 {
    color: #6a7aaa;
}

/* Comment.Special */
.pdoc .cs {
    color: #6a7aaa;
}

/* Generic.Deleted */
.pdoc .gd {
    color: #6a7aaa;
}

/* Generic.Emph */
.pdoc .ge {
    font-style: italic;
}

/* Generic.Inserted */
.pdoc .gi {
    color: #a6e22e;
}

/* Generic.Output */
.pdoc .go {
    color: #ff79c6;
}

/* Generic.Prompt */
.pdoc .gp {
    color: #f92672;
    font-weight: bold;
}

/* Generic.Strong */
.pdoc .gs {
    font-weight: bold;
}

/* Generic.Subheading */
.pdoc .gu {
    color: #75715e;
}

/* Keyword.Constant */
.pdoc .kc {
    color: #ff79c6;
}

/* Keyword.Declaration */
.pdoc .kd {
    color: #bd93f9;
}

/* Keyword.Namespace */
.pdoc .kn {
    color: #ff79c6;
}

/* Keyword.Pseudo */
.pdoc .kp {
    color: #bd93f9;
}

/* Keyword.Reserved */

.pdoc .kr {
    color: #ff79c6;
}

/* Keyword.Type */

.pdoc .kt {
    color: #ff79c6;
}

/* Literal.Date */

.pdoc .ld {
    color: #e6db74;
}

/* Literal.Number */

.pdoc .m {
    color: #ae81ff;
}

/* Literal.String */

.pdoc .s {
    color: #e6db74;
}

/* Name.Attribute */

.pdoc .na {
    color: #a6e22e;
}

/* Name.Builtin */

.pdoc .nb {
    color: #8be9fd;
}

/* Name.Class */

.pdoc .nc {
    color: #8be9fd;
}

/* Name.Constant */

.pdoc .no {
    color: #ff79c6;
}

/* Name.Decorator */

.pdoc .nd {
    color: #8be9fd;
}

/* Name.Entity */

.pdoc .ni {
    color: #ff79c6;
}

/* Name.Exception */

.pdoc .ne {
    color: #8be9fd;
}

/* Name.Function */

.pdoc .nf {
    color: #50fa7b;
}

/* Name.Label */

.pdoc .nl {
    color: #f8f8f2;
}

/* Name.Namespace */

.pdoc .nn {
    color: #f8f8f2;
}

/* Name.Other */

.pdoc .nx {
    color: #a6e22e;
}

/* Name.Property */

.pdoc .py {
    color: #f8f8f2;
}

/* Name.Tag */

.pdoc .nt {
    color: #f92672;
}

/* Name.Variable */

.pdoc .nv {
    color: #f8f8f2;
}

/* Operator.Word */

.pdoc .ow {
    color: #ff79c6;
}

/* Text.Whitespace */

.pdoc .w {
    color: #f8f8f2;
}

/* Literal.Number.Bin */

.pdoc .mb {
    color: #ae81ff;
}

/* Literal.Number.Float */

.pdoc .mf {
    color: #ae81ff;
}

/* Literal.Number.Hex */
.pdoc .mh {
    color: #ae81ff;
}

/* Literal.Number.Integer */
.pdoc .mi {
    color: #ae81ff;
}

/* Literal.Number.Oct */
.pdoc .mo {
    color: #ae81ff;
}

/* Literal.String.Affix */
.pdoc .sa {
    color: #ff79c6;
}

/* Literal.String.Backtick */
.pdoc .sb {
    color: #e6db74;
}

/* Literal.String.Char */
.pdoc .sc {
    color: #e6db74;
}

/* Literal.String.Delimiter */
.pdoc .dl {
    color: #e6db74;
}

/* Literal.String.Doc */
.pdoc .sd {
    color: #6272a4;
}

/* Literal.String.Double */
.pdoc .s2 {
    color: #e6db74;
}

/* Literal.String.Escape */
.pdoc .se {
    color: #ae81ff;
}

/* Literal.String.Heredoc */
.pdoc .sh {
    color: #e6db74;
}

/* Literal.String.Interpol. AKA f-strings */
.pdoc .si {
    color: #bd93f9;
}

/* Literal.String.Other */
.pdoc .sx {
    color: #e6db74;
}

/* Literal.String.Regex */
.pdoc .sr {
    color: #e6db74;
}

/* Literal.String.Single */
.pdoc .s1 {
    color: #e6db74;
}

/* Literal.String.Symbol */

.pdoc .ss {
    color: #e6db74;
}

/* Name.Builtin.Pseudo */
.pdoc .bp {
    color: #bd93f9;
}

/* Name.Function.Magic */
.pdoc .fm {
    color: #bd93f9;
}

/* Name.Variable.Class */
.pdoc .vc {
    color: #bd93f9;
}

/* Name.Variable.Global */
.pdoc .vg {
    color: #f8f8f2;
}

/* Name.Variable.Instance */
.pdoc .vi {
    color: #ffffff;
}

/* Name.Variable.Magic */
.pdoc .vm {
    color: #bd93f9;
}

/* Literal.Number.Integer.Long */
.pdoc .il {
    color: #ae81ff;
}

/* Decorator module i.e., @typing.? */
.pdoc .nd {
    color: #50fa7b;
}

/* This is for the actual doc style */

/* Font */
* {
    font-family: SFMono-Regular, Menlo, Monaco, Consolas, "Liberation Mono", "Courier New", monospace;
}

.pdoc .classattr {
    color: #fff;
}

/* Docstrings */

.pdoc p {
    color: #fff;
    font-weight: 300;
}

/* Module name */

.pdoc .modulename {
    color: #fff;
    font-weight: 0;
    margin: 0.3em 0;
    padding: 0.2em 0;
}

/* header */
.pdoc h5 {
    color: white;
    font-style: italic;
}

/* nav function */
nav.pdoc a.function {
    color: #4dec75;
}

/* nav vars. For some reason pdoc treats properties as class vars */
nav.pdoc a.variable {
    color: white;
}

/* Nav class name  */
nav.pdoc a.class {
    color: #8be9fd;
}

nav.pdoc a.class:hover {
    color: #fff;
}

/* nav modules */
nav.pdoc li a {
    color: #bd93f9;
}

/* Assigned values */
.pdoc span.default_value {
    color: #e6db74;
}

.pdoc .attr {
    color: white;
}
/*
Attributes
----------
*/

.pdoc h6#attributes {
    color: #fff;
    font-size: 23px;
}

/*
Example
-------
*/

.pdoc h6#example {
    color: #fff;
    font-size: 29px;
}

/*
Examples
--------
*/

.pdoc h6#examples {
    color: #fff;
    font-size: 29px;
}

/*
Notes
-----
*/

.pdoc h6#notes {
    font-size: 23px;
    color: orange;
}

/*
Warnings
-----
*/

.pdoc h6#warnings {
    font-size: 23px;
    color: orange;
}

/*
Returns
-------
*/

.pdoc h6#returns {
    font-size: 23px;
    color: white;
}

/*
Parameters
----------
*/

.pdoc h6#parameters {
    font-size: 23px;
    color: white;
}

/*
Other Parameters
----------
*/

.pdoc h6#other-parameters {
    font-size: 23px;
    color: white;
}

/*
Raises
------
*/

.pdoc h6#raises {
    font-size: 23px;
    color: #ff6666;
}

/* False, True literals */
.pdoc .kc {
    color: #bd93f9;
}

.pdoc li {
    color: white;
}

/* Decorator color */
.pdoc div.decorator {
    color: #50fa7b;
}

/* Function docs params name. i.e.

Parameters / Returns / Raises
----------
name : type
    A parameter
*/
.pdoc strong,
b {
    color: rgb(244, 90, 177);
    font-weight: normal;
}

/* colors for .. warning:: and .. note:: */
.pdoc em {
    color: orange;
}

/* Class name */
.pdoc span.name {
    color: #8be9fd;
}

/* Inherited class name */
.pdoc span.base {
    color: #8be9fd;
}

/* Functions */
.pdoc span.def {
    font-weight: normal;
}

/* Before inherited members colors, i.e., "builtins.Exception" */
.pdoc .inherited dt,
.pdoc .inherited dt::before {
    color: #8be9fd;
}

/* Whatever comes after the type inherited member. */
.pdoc .inherited dd,
.pdoc .inherited dd {
    color: #fff;
}

/* Commas that separates parameters "," color */
.pdoc .inherited dd:not(:last-child)::after {
    color: #fff;
}

/* Top left nav button */
nav.pdoc .module-list-button {
    display: inline-flex;
    align-items: center;
    margin-bottom: 1rem;
    color: white;
    border-color: white;
}

nav.pdoc .module-list-button:hover {
    border-color: white;
    color: white;
}

/* Contents, Submodules, API Documentations */
/* This also can be separated */
.pdoc h1,
.pdoc h2,
.pdoc h3 {
    font-weight: 300;
    margin: 0.3em 0;
    padding: 0.2em 0;
    color: white;
}
